import javax.realtime.*;

public class RecoverableAction implements ThreeWayRecoverableAtomicAction
{



  protected RecoverableController Control;
  private final boolean abort = false;
  private final boolean commit = true;

  private AtomicActionException aae1, aae2, aae3;

  public RecoverableAction()  // constructor
  {
    Control = new RecoverableController();
      // for recovery
      aae1 = new AtomicActionException();
      aae2 = new AtomicActionException();
      aae3 = new AtomicActionException();
  }

  class RecoverableController {
    protected boolean firstHere, secondHere, thirdHere;
    protected int allDone;
    protected int toExit, needed;
    protected int numberOfParticipants;
  private boolean committed = commit;
    //private AtomicActionException aae1, aae2, aae3;


    RecoverableController() 
    {
      // for synchronization
      firstHere = false;
      secondHere = false;
      thirdHere = false;
      allDone = 0;
      numberOfParticipants = 3;
      toExit = numberOfParticipants;
      needed = numberOfParticipants;

    }

    synchronized void first() throws InterruptedException
    {
      while(firstHere) wait();
      firstHere = true;
    }

    synchronized void second() throws InterruptedException
    {
      while(secondHere) wait();
      secondHere = true;
    }

    synchronized void third() throws InterruptedException
    {
      while(thirdHere) wait();
      thirdHere = true;
    }

    synchronized void signalAbort(Exception e)
    {
     allDone = 0;
     AtomicActionException.cause = e;
     AtomicActionException.wasInterrupted = true;
     aae1.fire();
     aae2.fire();
     aae3.fire();
    }

    private void reset() 
    {
      firstHere = false;
      secondHere = false;
      thirdHere = false;
      allDone = 0;
      toExit = numberOfParticipants;
      needed = numberOfParticipants;
      notifyAll();
    }

    synchronized void done() throws InterruptedException
    {
      allDone++;
      if(allDone == needed) {
        notifyAll();
      } else while(allDone != needed) {
           wait();  
      } 
      toExit--;
      if(toExit == 0) {
        reset();
      }
    }

    synchronized void cleanup(boolean abort)
    {
      if(abort) { committed = false; };
    }

    synchronized boolean waitCleanup() throws InterruptedException
    {
      allDone++;
      if(allDone == needed) {
        notifyAll();
      } else while(allDone != needed) {
           wait();  
      } 
      toExit--;
      if(toExit == 0) {
        reset();
      }
      return committed;
    };
  };
  public void role2() throws AtomicActionFailure { };
  public void role3() throws AtomicActionFailure { };
  public void role1() throws AtomicActionFailure { };
  public void rolex() throws AtomicActionFailure, AsynchronouslyInterruptedException
  {
    boolean Ok;
    // entry protocol
    // no aie until inside the atomic action
    boolean done = false;
    while(!done) {
      try {
        Control.first();
        done = true;
      } catch (InterruptedException e) {
        // ignore
      }
    } 
    
    Ok = aae1.doInterruptible
        (new Interruptible() 
          {
            public void run(AsynchronouslyInterruptedException e) 
                  throws AsynchronouslyInterruptedException 
            {
              try {

               // perform action
               // if necessary call e.disable() and e.enable() to
               // defer aie
               Control.done();
              }
              catch(Exception x) {
                if(x.getClass().getName() == "AsynchronouslyInterruptedException")
                  ((AsynchronouslyInterruptedException) x).propagate();
                else
                  Control.signalAbort(x);
              }
            }

            public void interruptAction(
                  AsynchronouslyInterruptedException e)
            {
              AtomicActionException Myaae = (AtomicActionException) e;
              // Myaae.cause is the cause of the problem
            }
          }
        );
      if(!Ok) throw new AtomicActionFailure();
      if(aae1.wasInterrupted) {
        try {
          // try to recover
          Control.cleanup(commit);
          if(Control.waitCleanup() != commit) {
             throw new AtomicActionFailure();              
          };
        }
        catch(Exception x) {
          throw new AtomicActionFailure(); 
        }
      };
    }
  }
